////////////////////////////////////////////////////////////////
// April 2012. (C) Paragon Software
// Valery Kovalev
//
// This file contain declaration and definitions of
// Compress IO interface.
// This file was created from "./imagemgm/inc/vimdisk.h"
////////////////////////////////////////////////////////////////


#ifndef __UFSD_COMPRESS_H
#define __UFSD_COMPRESS_H


namespace UFSD {


//Compression algorithms
#define I_COMPRESS_NONE         0                       /* without compression */
#define I_COMPRESS_DEFLATE      1                       /* DEFLATE (RFC 1951) */
#define I_COMPRESS_LZ4          2                       /* http://code.google.com/p/lz4/ */
#define I_COMPRESS_RLEP         3                       /* RLE paragon modificated */

///////////////////////////////////////////////
//
//   Compression
//
///////////////////////////////////////////////

// provider
class ICompress
{
public:

  virtual unsigned    SetCompressLevel(unsigned int CompressLevel) = 0; //0-default compress level (fast)

  virtual unsigned    Compress(
    const void*         in_buf,
    unsigned int        in_len,
    void*               out_buf,
    unsigned int        out_max,
    unsigned int*       out_len
  ) = 0;

  virtual unsigned    Decompress(
    const void*         in_buf,
    unsigned int        in_len,
    void*               out_buf,
    unsigned int        out_max,
    unsigned int*       out_len
  ) = 0;

  virtual void        Destroy() = 0;
};


// factory
class ICompressIo
{
public:

  virtual unsigned    CreateProvider(
    unsigned int        Method,                         //I_COMPRESS_XXX
    ICompress**         ppICompress
  ) = 0;
};


} //namespace UFSD


#endif //__UFSD_COMPRESS_H


