////////////////////////////////////////////////////////////////
// April 2012. (C) Paragon Software
// Valery Kovalev
//
// This file contain declaration and definitions of
// Crypt IO interface.
////////////////////////////////////////////////////////////////


#ifndef __UFSD_CRYPT_H
#define __UFSD_CRYPT_H


namespace UFSD {


//Crypt algorithms
#define I_CRYPT_NONE            0                       /* without cryption */
#define I_CRYPT_OLHA            1                       /* crypt Olha */

#define MIN_KEY_BLOB_LENGTH     100

///////////////////////////////////////////////
//
//   Cryptions
//
///////////////////////////////////////////////

// provider
class ICrypt
{
public:

  //Generate new operation key
  virtual unsigned      GenKey(
    const void*           pPwd,                         //Password
    unsigned int          PwdLen,
    unsigned int          KeyLen = 0,                   //0-default key len (minimal)
    unsigned int          Rnd = 0                       //initial random number, 0-not changed
  ) = 0;

  //Change password
  virtual unsigned      ChangePwd(
    const void*           pPwd,                         //Password
    unsigned int          PwdLen
  ) = 0;

  //Check and set operation key
  virtual unsigned      SetBlobKey(
    const void*           pPwd,                         //Password
    unsigned int          PwdLen,
    const void*           pBlobKeyBuff,                 //Encrypted key
    unsigned int          BlobKeyLen
  ) = 0;

  //Get Blob operation key
  virtual unsigned      GetBlobKey(
    void*                 pBlobKeyBuff,                 //Returned encrypted key
    unsigned int          BuffLen,
    unsigned int*         pBlobKeyLen
  ) = 0;

  virtual unsigned      Encrypt(
    void*                 pBuff,
    unsigned int          BuffLen,
    unsigned int*         pDataLen                      //crypted and encripted size can be differ
  ) = 0;

  virtual unsigned      Decrypt(
    void*                 pBuff,
    unsigned int          BuffLen,
    unsigned int*         pDataLen                      //crypted and encripted size can be differ
  ) = 0;

  unsigned              MakeBlobKey(
    const void*           pPwd,                         //Password
    unsigned int          PwdLen,
    void*                 pBlobKeyBuff,                 //Encrypted key
    unsigned int          BlobKeyLen
  )
  {
    unsigned rc = GenKey(pPwd, PwdLen, BlobKeyLen);
    if(0 != rc)
      return rc;

    return GetBlobKey(pBlobKeyBuff, BlobKeyLen, NULL);
  }

  unsigned              CheckBlobKey(
    const void*           pPwd,                         //Password
    unsigned int          PwdLen,
    const void*           pBlobKeyBuff,                 //Encrypted key
    unsigned int          BlobKeyLen
  )
  {
    return SetBlobKey(pPwd, PwdLen, pBlobKeyBuff, BlobKeyLen);
  }

  virtual void          Destroy() = 0;
};


// factory
class ICryptIo
{
public:

  virtual unsigned      CreateProvider(
    unsigned int          Method,                       //I_CRYPT_XXX
    ICrypt**              ppICrypt
  ) = 0;
};


} //namespace UFSD


#endif //__UFSD_CRYPT_H
