////////////////////////////////////////////////////////////////
// October 2012. (C) Paragon Software
// Valery Kovalev
//
// This file contain declaration and definitions of
// File-Blocks deduplication interface.
////////////////////////////////////////////////////////////////


#ifndef __UFSD_BDEDUP_H
#define __UFSD_BDEDUP_H


namespace UFSD {


#ifndef NO_PRAGMA_PACK
#pragma pack( push, 1 )
#endif


struct BlockID
{
  //
  UINT64        BlockOffset:48;   // 6 bytes block offset in 512 bytes from begin of file
  UINT64        BlockSize:14;     // size of not zero data in block in 512 bytes
  UINT64        Flags:2;          // some additional block info
  unsigned int  Crc;              // cyclic hash of block
  unsigned char Hash[20];         // hash of block 160 bit max
};

#ifdef C_ASSERT
C_ASSERT(sizeof(BlockID) == 32);
#endif


#ifndef NO_PRAGMA_PACK
#pragma pack( pop )
#endif


class IFileBlockDedupServer
{
public:

  // return deduplication data block size
  virtual unsigned int  GetBlockSize() = 0;
  // minimal data size for deduplication
  virtual unsigned int  GetMinDedupSize() = 0;

  // caller fills pID->BlockOffset and pData
  // server fills all other fields in pID and saves block data if need
  virtual unsigned      PutBlocks(BlockID* pId, unsigned int Count, void* pData) = 0;
  // server returns block data for specified ID
  virtual unsigned      GetBlocks(BlockID* pId, unsigned int Count, void* pData) = 0;
  // query blocks for statistic and fast restor
  virtual unsigned      QueryBlocks(BlockID* pId, unsigned int Count) = 0;

  // Only files for deduplication
  // if specified all other files will not be processed
  virtual unsigned      GetDedupList(STRType* pType, void** pIncludes) = 0;

  // Exclude files from deduplication
  virtual unsigned      GetExcludeList(STRType* pType, void** pExcludes) = 0;

  virtual unsigned      FreeList(void* pList) = 0;
};


} //namespace UFSD


#endif //__UFSD_BDEDUP_H
