/****************************************************
 *  File UFSD/include/errors.h
 *  Copyright (C) 2000 Paragon Software,
 *
 *  Error codes are listed here.
 ****************************************************/

#ifndef __UFSD_ERRORS_H
#define __UFSD_ERRORS_H

#ifndef UFSD_ERROR_BASE
  // Severity bit, Customer bit
  #define UFSD_ERROR_BASE  0xA0000100
#endif

#define MAKE_UFSD_ERROR(x)  ( UFSD_ERROR_BASE | (x) )
#define IS_UFSD_ERROR(x)    ( UFSD_ERROR_BASE == (UFSD_ERROR_BASE & (x) ) )



#define ERR_NOERROR         0x0000


#define ERR_BADPARAMS       MAKE_UFSD_ERROR(0x0001)
#define ERR_NOMEMMNGR       MAKE_UFSD_ERROR(0x0002)
#define ERR_BADRWBLOCK      MAKE_UFSD_ERROR(0x0003)
#define ERR_FSUNKNOWN       MAKE_UFSD_ERROR(0x0006)
#define ERR_NOMEMORY        MAKE_UFSD_ERROR(0x0007)
#define ERR_BADUFSDFILE     MAKE_UFSD_ERROR(0x0008)
#define ERR_OPENROOT        MAKE_UFSD_ERROR(0x0009)
#define ERR_OPENFILE        MAKE_UFSD_ERROR(0x000A)
#define ERR_CREATEFILE      MAKE_UFSD_ERROR(0x000B)
#define ERR_READFILE        MAKE_UFSD_ERROR(0x000C)
#define ERR_WRITEFILE       MAKE_UFSD_ERROR(0x000D)
#define ERR_NOFILEEXISTS    MAKE_UFSD_ERROR(0x000E)
#define ERR_FILEEXISTS      MAKE_UFSD_ERROR(0x000F)
#define ERR_NOROOTDIR       MAKE_UFSD_ERROR(0x0010)
#define ERR_NOUFSDINIT      MAKE_UFSD_ERROR(0x0011)
#define ERR_NOCURRENTDIR    MAKE_UFSD_ERROR(0x0012)
#define ERR_BADNAME         MAKE_UFSD_ERROR(0x0013)
#define ERR_BADNAME_LEN     206   // ERROR_FILENAME_EXCED_RANGE
#define ERR_BADSTRTYPE      MAKE_UFSD_ERROR(0x0014)
#define ERR_CLOSEUNABLE     MAKE_UFSD_ERROR(0x0016)
#define ERR_NODIRENTRY      MAKE_UFSD_ERROR(0x0017)
#define ERR_BADOBJECTTYPE   MAKE_UFSD_ERROR(0x0019)
#define ERR_CREATEDIR       MAKE_UFSD_ERROR(0x001B)
#define ERR_NOFSINTEGRITY   MAKE_UFSD_ERROR(0x001C)
//#define ERR_DELOPENFILE     MAKE_UFSD_ERROR(0x001D)
//#define ERR_DELOPENDIR      MAKE_UFSD_ERROR(0x001E)
#define ERR_BADATTRIBUTES   MAKE_UFSD_ERROR(0x001F)
#define ERR_WPROTECT        MAKE_UFSD_ERROR(0x0020)
#define ERR_DIRNOTEMPTY     MAKE_UFSD_ERROR(0x0021)
#define ERR_FILENODATA      MAKE_UFSD_ERROR(0x0022)
// Not enough free space
#define ERR_NOSPC           MAKE_UFSD_ERROR(0x0023)

// Errors used in IoControl:
// This function is not implemented
#define ERR_NOTIMPLEMENTED      MAKE_UFSD_ERROR(0x0024)
// If the output buffer is too small to return any data
#define ERR_INSUFFICIENT_BUFFER MAKE_UFSD_ERROR(0x0025)
// If the output buffer is too small to hold all of the data but can hold some entries
#define ERR_MORE_DATA           MAKE_UFSD_ERROR(0x0026)


// Read/write bad block
#define ERR_BADBLOCK            MAKE_UFSD_ERROR( 0x0027)
// User cancels operation
#define ERR_CANCELLED           MAKE_UFSD_ERROR( 0x0028)

// Wrong or not supported object format
#define ERR_WRONGFORMAT         MAKE_UFSD_ERROR(0x0029)
// Wrong data from ZIP archive data may be encrypted
#define ERR_WRONGDATA           MAKE_UFSD_ERROR(0x002A)

// Can't mount 'cause journal is not empty
#define ERR_NEED_REPLAY         MAKE_UFSD_ERROR(0x002B)

#define ERR_LOG_FULL            MAKE_UFSD_ERROR(0x002C)

// IChildThread errors
// thread already running
#define ERR_ALREADY_RUNNING     MAKE_UFSD_ERROR(0x002D)
// error while create thread
#define ERR_CREATE_THREAD       MAKE_UFSD_ERROR(0x002E)

// IRestartSaver errors
// error while start saving
#define ERR_START_SAVER         MAKE_UFSD_ERROR(0x002F)
// saver not started
#define ERR_SAVER_NOTSTARTED    MAKE_UFSD_ERROR(0x0030)

// Volumes are different and cannot be compared
#define ERR_VOLUMES_INCOMPATIBLE MAKE_UFSD_ERROR(0x0031)

// User aborts operation
#define ERR_UFSD_CANCELLED      MAKE_UFSD_ERROR( 0x0032)

// Media is ejected while operation
#define ERR_NOMEDIA             MAKE_UFSD_ERROR( 0x0033 )

// Generic not success
#define ERR_NOT_SUCCESS         MAKE_UFSD_ERROR( 0x0034 )

// Generic error due to failed encryption\decryption operation
#define ERR_ENCRYPTION	        MAKE_UFSD_ERROR( 0x0035 )

// Lock required to perform operation
#define ERR_LOCK_REQ            MAKE_UFSD_ERROR( 0x0036 )

// Need to restart log
#define ERR_LOG_RESTART         MAKE_UFSD_ERROR( 0x0037 )

#define ERR_UNKNOWN             0xFFFFFFFFu

//
// Use this define to simplify error checking
//
#define UFSD_SUCCESS(x) ( ERR_NOERROR == (x) )



#endif //__UFSD_ERRORS_H

