////////////////////////////////////////////////////////////////
// April 2012. (C) Paragon Software
// Valery Kovalev
//
// This file contain declaration and definitions of
// File IO interface.
////////////////////////////////////////////////////////////////


#ifndef __UFSD_FILEIO_H
#define __UFSD_FILEIO_H


namespace UFSD {


//File mode
#define I_FILE_READ                     0x01            // read
#define I_FILE_WRITE                    0x02            // write
#define I_FILE_SHARE_READ               0x04            // share read
#define I_FILE_SHARE_WRITE              0x08            // share write

//flags
#define I_FILE_SPARSE                   0x10000         // create as sparse file

//Creation disposition
#define I_FILE_CREATE                   0x01            // create truncated
#define I_FILE_OPEN                     0x02            // open existing

///////////////////////////////////////////////
//
//   File  i/o
//
///////////////////////////////////////////////
//all errors UFSD_ERROR

// file
class IFile
{
public:

   // destroy object
   virtual unsigned   Close()                           = 0;

   // flush file
   virtual unsigned   Flush()                           = 0;

   // get file size in bytes
   virtual unsigned   GetSize(UINT64* cSize)            = 0;

   // set file size in bytes
   virtual unsigned   SetSize(UINT64  newSize)          = 0;

   // read file bytes from offset
   virtual unsigned   ReadBytes(
     UINT64             Start,
     void*              Buf,
     unsigned int       Len)                            = 0;

   // write file bytes from offset
   virtual unsigned   WriteBytes(
     UINT64             Start,
     const void*        Buf,
     unsigned int       Len)                            = 0;

   // get file times
   virtual unsigned   GetTimes(
     UINT64*            CrTime,
     UINT64*            ModTime,
     UINT64*            ReffTime)                       = 0;
};


// factory
class IFileIo
{
public:

   // open or create file
   virtual unsigned   OpenCreate(
     const char*        Path,                                   //utf-8
     unsigned int       FileMode,                               //read/write/share_read/share_write
     unsigned int       DispositionAndFlags,                    //open/create
     IFile**            pFile)                           = 0;

   // delete file
   virtual unsigned   Delete(const char* Path)           = 0;   //utf-8

   // check if file exists
   virtual unsigned   Exists(
     const char*        Path,
     int*               Ret)                             = 0;   //utf-8

   // make directory
   virtual unsigned   MkDir(const char* Path)            = 0;   //utf-8
};


} //namespace UFSD


#endif //__UFSD_FILEIO_H

