////////////////////////////////////////////////////////////////
// Feb 2012. (C) Paragon Software
// Valery Kovalev
//
// This file contain Index File specification
// This is parts of ITE
//
////////////////////////////////////////////////////////////////


#ifndef __U_INDEXF_H__
#define __U_INDEXF_H__

namespace UFSD {


#ifndef NO_PRAGMA_PACK
#pragma pack( push, 1 )
#endif


////////////////////////////////////////////////////////////////
#define BLOCK_SIGN_ARCINFO              0x3a696150 /*"Pai:"*/
#define BLOCK_SIGN_ARCEND               0x3a656150 /*"Pae:"*/
#define BLOCK_SIGN_PADDING              0x3a647050 /*"Ppd:"*/
#define BLOCK_SIGN_FILENAME             0x3a6e6650 /*"Pfn:"*/
#define BLOCK_SIGN_FILEINFO             0x3a696650 /*"Pfi:"*/
#define BLOCK_SIGN_FILEATTRIBUTES       0x3a616650 /*"Pfa:"*/
#define BLOCK_SIGN_FILEDATA             0x3a646650 /*"Pfd:"*/

#define BLOCK_PVHD_DISK_INFO            ':idP' /*"Pdi:"*/
#define BLOCK_PVHD_EXTENT_INFO          ':ieP' /*"Pei:"*/
#define BLOCK_PVHD_EXTENT_DESC          ':deP' /*"Ped:"*/
#define BLOCK_PVHD_BLOCK_DESC           ':dbP' /*"Pbd:"*/
#define BLOCK_PVHD_EXTENT_ADDRESS       ':aeP' /*"Pea:"*/
#define BLOCK_PVHD_PARTS_INFO           ':ipP' /*"Ppi:"*/
#define BLOCK_PVHD_DISK_COMMENT         ':cdP' /*"Pdc:"*/

#define BLOCK_DVHD_DISK_INFO            ':idD' /*"Ddi:"*/
#define BLOCK_DVHD_BLOCK_ADDRESS        ':abD' /*"Dba:"*/


#define MAX_BLOCKSIZE 0x800000

struct BlockHeader
{
  unsigned int          Sign;           // signature for identify block type
  unsigned int          BlockSize;      // size of current block with this header
  unsigned int          Index;          // some block number counted from 1
  unsigned int          RecordSize;     // size of one record, if 0 records have variable size

  unsigned int          UsedSize;       // used size in this block without this header
  unsigned int          Reserved;
  union {
  UINT64                Next64;
  BlockHeader*          Next;           // link to next block in memory
  struct {
    unsigned int        Header;         // crc of previous field in header
    unsigned int        Data;           // crc of block data with UsedSize
    } Crc;
  };
  // block data
  // array of records
};
#ifdef C_ASSERT
C_ASSERT(sizeof(BlockHeader) == 32);
#endif

#define CalcBlockSize(used_size)          (((used_size) + sizeof(UFSD::BlockHeader) + 0xf) & ~0xf)
#define CalcBlockSizeEx(used_size, align) (((used_size) + sizeof(UFSD::BlockHeader) + (align - 1)) & ~(align - 1))

struct RecordHeader
{
  unsigned short        Type;           // type of record
  unsigned short        Reserved;
  unsigned int          RecordSize;     // size of this record with this header
  union {
  UINT64                NextIndex;      // next record index in list
  RecordHeader*         Next;           // link to next record
  };
  // record data
};
#ifdef C_ASSERT
C_ASSERT(sizeof(RecordHeader) == 16);
#endif

////////////////////////////////////////////////////////////////
// API for working with Index File content
// Works only in READ or WRITE mode
class UFSD_ABSTRACT_CLASS CIndexFile : public CMemBased<CIndexFile>
{
public:
  CIndexFile(
    IN CMemoryManager*  mm
  )
  : CMemBased<CIndexFile>(mm)
  {}

  virtual void Destroy() = 0;

  // flush all data
  // call this before Destroy() if you to want check writting all data to disk
  virtual unsigned Flush() = 0;

  // return the begin of current block position in file
  virtual UINT64 GetCurPos() = 0;

  // reset index file position
  // for read only mode
  virtual unsigned Rewind() = 0;

  // find specified block (BlockSignature) from current position
  // we can find not only first block (SkipBlock parameter)
  virtual unsigned FindBlock(unsigned int BlockSignatute, unsigned int SkipBlocks, size_t* pBlockSize) = 0;

  // get current block header
  virtual unsigned GetBlockHeader(BlockHeader* Header) = 0;

  // read next block
  virtual unsigned ReadBlock(void* Buffer, size_t BufferSize, size_t* Read) = 0;

  // write next block to end fo file
  // written block will be aligned
  // intermediate block aligned at 16 Bytes
  // last block aligned at 16 KBytes edge
  virtual unsigned AppendBlock(void* Buffer, size_t BufferSize, bool fLastBlock, size_t* Written) = 0;

  // change current block data
  // header of block must be same (not changed)
  // after writting current position will point to next block in file
  virtual unsigned ChangeBlock(void* Buffer, size_t BufferSize, size_t* Written) = 0;
};


#ifndef NO_PRAGMA_PACK
#pragma pack( pop )
#endif

} //namespace UFSD

#endif //__U_INDEXF_H__
