////////////////////////////////////////////////////////////////
// May of 2004. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//

#ifndef __UFSD_MSG_H
#define __UFSD_MSG_H

namespace UFSD{


/*
*---------------------------------------------------------*
*                 Progress dialog                         *
*---------------------------------------------------------*
* |---------------------------------------------------|   *
* |  Log messages goes here                           |   *
* |  scrollable window                                |   *
* |                                                   |   *
* |---------------------------------------------------|   *
*                                                         *
*                                                         *
* |---------------------------------------------------|   *
* |  rewritable messages goes here                    |   *
* |---------------------------------------------------|   *
*                                                         *
*---------------------------------------------------------*

  Typical sequence of operation:

  - OpenBar();
  - PutMessage( Title, "Test" );

  - NewProgress( Max1, "Cycle 1" )
  for ( i = 0; i < Max1; ++i )
  {
    - SetPos( i );
    - PutMessage( Cur, "%d", i );
  }

  - NewProgress( Max2, " Cycle 2")
  for ( i = 0; i < Max2; ++i )
  {
    - SetPos( i );
    - PutMessage( Cur, "%d", i );
  }

  - CloseBar();


*/


// Possible values for first argument of PutMessage
enum MsgType{
  MsgTitle,     // Set new title
  MsgLog,       // Put message in scrollable window
  MsgCur        // Put message in rewritable window
};

struct UFSD_ABSTRACT_CLASS IMessage
{
  //=======================================================
  //           Progress bar related functions
  //=======================================================
  // Open progress bar.
  // Returns 0 if error
  virtual int OpenBar(
      IN const char*  szTitle, ...
      ) __attribute__ ((format (printf, 2, 3))) = 0;

  // Close progress bar.
  // Returns 0 if error
  virtual int CloseBar() = 0;

  // Starts a new progress bar with given maximum value of position
  virtual int NewProgress(
      IN size_t         Flags,      // see UMSG_FLAGS_XXX
      IN const UINT64&  MaxValue,
      IN const char*    szTitle, ...
      )  __attribute__ ((format (printf, 4, 5))) = 0;

  // Gets the upper limits of the range of the progress bar
  virtual UINT64 GetMax(
      ) = 0;

  // Sets the current position for a progress bar
  virtual void  SetPos(
      IN const UINT64& Pos
      ) = 0;

  // Gets the current position of the progress bar
  virtual UINT64 GetPos(
      ) = 0;

  // Advances the current position for a progress bar by the step increment
  virtual void  StepIt(
      IN const UINT64& Step = 1
      ) = 0;

  // Returns current percent value
  virtual size_t GetPercent() = 0;


  //=======================================================
  //           Messages functions
  //=======================================================

  // Put formatted message
  // Returns the number of put chars
  // NOTE: "this" is passed implicitly as a first parameter
  // so the "fmt" will be actually the third parameter
  virtual int PutMessage(
      IN MsgType      type,
      IN const char*  fmt, ...
      ) __attribute__ ((format (printf, 3, 4))) = 0;


  //=======================================================
  //  These "callback" functions allow to ask user
  //=======================================================
  // Show message box
  // Return value 0 means that request is not implemented or is not processed
  virtual size_t MessageBox(
      IN size_t         flags,
      IN const char*    fmt, ...
      ) __attribute__ ((format (printf, 3, 4)) ) = 0;

  // Call synchronous request
  // The return value depends on "code"
  // Return value 0 means that request is not implemented or is not processed
  virtual size_t SyncRequest(
      IN size_t     code,
      IN OUT void*  arg
      ) = 0;

  //=======================================================
  // During long operations user can press "Cancel" button
  // This function returns TRUE in this case
  //=======================================================
  virtual bool  IsBreak(
      IN bool bClear = false
      ) = 0;
};


//
// IMessage::NewProgress() flags
//
#define UMSG_FLAGS_MAIN                 1 // This is the main subprogress
#define UMSG_ITE_COPY_DATA              2 // ITE operation copy data

//
// IMessage::MessageBox() flags
// (NOTE: you can pass this flag in Win32 MessageBox) as is
//

#define UMSG_MB_OK                       0x00000000
#define UMSG_MB_OKCANCEL                 0x00000001
#define UMSG_MB_ABORTRETRYIGNORE         0x00000002
#define UMSG_MB_YESNOCANCEL              0x00000003
#define UMSG_MB_YESNO                    0x00000004
#define UMSG_MB_RETRYCANCEL              0x00000005


//
// Predefined returned values of IMessage::MessageBox
//

#define UMSG_IDOK                1
#define UMSG_IDCANCEL            2
#define UMSG_IDABORT             3
#define UMSG_IDRETRY             4
#define UMSG_IDIGNORE            5
#define UMSG_IDYES               6
#define UMSG_IDNO                7


} //namespace UFSD{

#endif // __UFSD_MSG_H

