////////////////////////////////////////////////////////////////
// January 2010. (C) Paragon Software
// Valery Kovalev
//
// This file contain declaration and definitions of
// Restart Saver interface.
// This interface must be used for long FS operation with
// critical changing on disk structures
////////////////////////////////////////////////////////////////


#ifndef __UFSD_RSAVER_H
#define __UFSD_RSAVER_H


namespace UFSD {


//////////////////////////////////////////////////////////////
// Restart Saver interface
struct UFSD_ABSTRACT_CLASS IRestartSaver
{
public:
  //////////////////////////////////////////////////////////////
  // Restart Saver control functions
  //////////////////////////////////////////////////////////////

  // Check is Restart Saver active
  virtual bool     IsRunning() = 0;

  // Activate Restart Saver
  // On this command Restart Saver must:
  //  - increase number of activations, if first:
  //  - save and change MBR
  //  - lookup for maximum empty data space
  virtual unsigned Start() = 0;

  // Deactivate Restart Saver
  // On this command Restart Saver must:
  //  - decrease number of activations, if zero:
  //  - free any used data
  //  - restore MBR
  virtual unsigned Stop() = 0;


  //////////////////////////////////////////////////////////////
  // Restart Data access functions
  //////////////////////////////////////////////////////////////

  // Query maximal data space that can be used for saving the operation data
  virtual unsigned QueryDataSpace(
    OUT UINT64 *        SizeInBytes
  ) = 0;

  // Allocate specified size from data space
  virtual unsigned AllocData(
    IN const UINT64&    Bytes
  ) = 0;

  // Free previously allocated data
  virtual unsigned FreeData() = 0;

  // Get allocated data size
  // After restarting we can get saved size
  virtual unsigned GetDataSize(
    OUT UINT64 *        Bytes
  ) = 0;

  // Read data from allocated arrea
  // After restarting we can read saved data
  virtual unsigned ReadData(
    IN const UINT64&    OffsetInBytes,  // offset from begin of allocated arrea
    IN size_t           Bytes,          // bytes to read
    OUT void *          pBuff
  ) = 0;

  // Write data to allocated arrea
  virtual unsigned WriteData(
    IN const UINT64&    OffsetInBytes,  // offset from begin of allocated arrea
    IN size_t           Bytes,          // bytes to write
    IN const void *     pBuff
  ) = 0;

  //////////////////////////////////////////////////////////////
  // Fast counter (saved in CMOS or FLASH) access functions
  //////////////////////////////////////////////////////////////

  // Read fast counter
  // If counter unsupported value must be set to 0 and return specified error code
  virtual unsigned ReadCounter(
    OUT unsigned char * value
  )
    {if(value) *value = 0; return ERR_NOTIMPLEMENTED;}

  // Write fast counter
  // If counter unsupported return specified error code
  virtual unsigned WriteCounter(
    IN unsigned char    value
  )
    {(void)value; return ERR_NOTIMPLEMENTED;}
};


} //namespace UFSD


#endif //__UFSD_RSAVER_H
