////////////////////////////////////////////////////////////////
// February 2010. (C) Paragon Software
// Valery Kovalev
//
// This file contain declaration and definitions of
// Thread and Synchronization Objects interface.
////////////////////////////////////////////////////////////////


#ifndef __ITHREAD_H__
#define __ITHREAD_H__

#include "u_mmngr.h"


namespace UFSD {


/////////////////////////////////////////////////////////////////////////////
typedef unsigned (*ProcFunc)(void* pArg);

class IChildThread : public CMemBased<IChildThread>
{
public:
  IChildThread(CMemoryManager* Mm) : CMemBased<IChildThread>(Mm) {}
  virtual void     Destroy() = 0;

  virtual unsigned Start(ProcFunc pFunc, void* pArg) = 0;
  virtual void     Terminate() = 0;
};


/////////////////////////////////////////////////////////////////////////////
class IMutex : public CMemBased<IMutex>
{
public:
  IMutex(CMemoryManager* Mm) : CMemBased<IMutex>(Mm) {}
  virtual void     Destroy() = 0;

  virtual int      Lock() = 0;
  virtual int      Unlock() = 0;
};


/////////////////////////////////////////////////////////////////////////////
class IEvent : public CMemBased<IEvent>
{
public:
  IEvent(CMemoryManager* Mm) : CMemBased<IEvent>(Mm) {}
  virtual void     Destroy() = 0;

  virtual int      SetEvent() = 0;
  virtual int      ResetEvent() = 0;
  virtual int      WaitEvent() = 0;
};


/////////////////////////////////////////////////////////////////////////////
class IUThread : public CMemBased<IUThread>
{
public:
  IUThread(CMemoryManager* Mm) : CMemBased<IUThread>(Mm) {}

  virtual IChildThread* CreateIThread() = 0;
  virtual IMutex*       CreateIMutex() = 0;
  virtual IEvent*       CreateIEvent() = 0;

  virtual void          Sleep(unsigned int Milliseconds) = 0;
};


} //namespace UFSD

#endif //__ITHREAD_H__

