#ifndef __UFSD_TIME_H
#define __UFSD_TIME_H

namespace UFSD {

// UTC means Universal Coordinated Time
// UTC = local time + bias
class UFSD_ABSTRACT_CLASS CTime
{
protected:
  // Make compiler happy:
  CTime(){}
  CTime(const CTime&);
  CTime& operator=(const CTime&);

public:
    // This function returns UTC time:
    // 64-bit value representing the number of 100-nanosecond intervals since January 1, 1601
    virtual UINT64 Time() = 0;

    // The bias is the difference, in minutes, between UTC time and local time.
    // UTC = Local + Bias
    //
    // F.e.
    //  Eastern time zone: +300
    //  Paris,Berlin:      -60
    //  Moscow:            -180
    virtual long GetBias() = 0;

    // Number of 100ns in one second
    enum{ TicksPerSecond = 10*1000*1000 };
};

// Seconds since 1601
#define SecondsToStartOf1904 PU64(0x0000000239EAE080)
#define SecondsToStartOf1970 PU64(0x00000002B6109100)
#define SecondsToStartOf1980 PU64(0x00000002C8DF3700)

} // namespace UFSD

#endif //__UFSD_TIME_H

