////////////////////////////////////////////////////////////////
// March of 2005. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file contains declarations of many external functions
// used in this project
//
// NOTE: File "ufsd.h" should be included before this file
////////////////////////////////////////////////////////////////

#ifndef _UFSD_FUNCS_H
#define _UFSD_FUNCS_H

#ifndef IN
# define IN
# define OUT
#endif

#ifndef PACKAGE_TAG
  #define PACKAGE_TAG "(" __DATE__ " " __TIME__ ")"
#endif

#define VER_STRING _(NLS_CAT_NAME " " PACKAGE_TAG "\nParagon Software Group\n\n")


#ifdef __cplusplus
#define EXTERN_C extern "C"
#else
#define EXTERN_C extern
#endif


///////////////////////////////////////////////////////////
// UFSD_GetTimeService
//
// Returns time manager to be used in UFSD. See ufsdtime.cpp
///////////////////////////////////////////////////////////
UFSD::CTime*
UFSD_GetTimeService();

///////////////////////////////////////////////////////////
// UFSD_GetStringsService
//
// Returns string manager to be used in UFSD. See ufsdstr.cpp
///////////////////////////////////////////////////////////
UFSD::CStrings*
UFSD_GetStringsService();

///////////////////////////////////////////////////////////
// UFSD_GetMemoryManager
//
// Returns memory manager to be used in UFSD. See ufsdmmgr.cpp
///////////////////////////////////////////////////////////
UFSD::CMemoryManager*
UFSD_GetMemoryManager();

///////////////////////////////////////////////////////////
// UFSD_IOHandlerCreate
//
// Returns read/write manager to be used in UFSD. See ufsdio.cpp
///////////////////////////////////////////////////////////
unsigned
UFSD_IOHandlerCreate(
    IN const char*        szDevice,
    IN bool               bReadOnly,
    OUT UFSD::CRWBlock**  Rw,
    OUT int*              fd  = NULL,
    IN bool               bForceDismount = false,
    IN bool               bVerbose = false
    );

///////////////////////////////////////////////////////////
// GetRwBlock
//
// Used to create rw block from fsdump file
///////////////////////////////////////////////////////////
UFSD::CRWBlock*
GetRwBlock(
    IN const char* szFileName
    );

///////////////////////////////////////////////////////////
// UFSD_IOHandlerFree
//
// Frees read/write manager allocated by UFSD_IOHandlerCreate.
///////////////////////////////////////////////////////////
void
UFSD_IOHandlerFree(
    IN UFSD::CRWBlock* p
    );

///////////////////////////////////////////////////////////
// UFSD_GetMessageService
//
// Returns message manager to be used in UFSD. See ufsdmsg.cpp
///////////////////////////////////////////////////////////
UFSD::IMessage*
UFSD_GetMessageService(
    IN FILE* fd = stdout,
    IN size_t (__stdcall *ExtSyncRequest)(
      IN UFSD::IMessage*  This,
      IN size_t           code,
      IN OUT void*        arg
      ) = NULL,
    IN bool bSkipPercents = false
    );

///////////////////////////////////////////////////////////
// utf8_wcstombs
//
// Converts a wide string to the corresponding multibyte string
///////////////////////////////////////////////////////////
int
utf8_wcstombs(char*s, int n, const unsigned short* pwcs, int m );

///////////////////////////////////////////////////////////
// EnableBreak
//
// Enable/Disable break
///////////////////////////////////////////////////////////
void
EnableBreak(
    IN bool bEnable
    );

///////////////////////////////////////////////////////////
// GetUfsdErrorStr
//
// Translates UFSD error code to string. See debug.c
///////////////////////////////////////////////////////////
EXTERN_C
const char*
GetUfsdErrorStr(
    IN unsigned nError
    );

///////////////////////////////////////////////////////////
// PrintUfsdStatusIfError
//
// Prints on stderr error string
///////////////////////////////////////////////////////////
EXTERN_C void
PrintUfsdStatusIfError(
    IN unsigned nError
    );


///////////////////////////////////////////////////////////
// EnableTrace
//
// Enables trace from UFSD
///////////////////////////////////////////////////////////
EXTERN_C
void
EnableTrace();

///////////////////////////////////////////////////////////
// CloseTrace
//
// This function closes trace file
///////////////////////////////////////////////////////////
EXTERN_C
void
CloseTrace();

///////////////////////////////////////////////////////////
// InitUfsdLibrary
//
// Inits libufsd.so
///////////////////////////////////////////////////////////
EXTERN_C
void
InitUfsdLibrary( void );

///////////////////////////////////////////////////////////
// ParseSize
//
// Parses string for size like
// 512, 1024, 2048, 4096, 8192, 16K, 32K, 64K, 32M
// Returns 0 if error
///////////////////////////////////////////////////////////
unsigned int
ParseSize(
    IN const char* s
    );

#endif // #ifndef _UFSD_FUNCS_H
