////////////////////////////////////////////////////////////////
// March of 2005. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file implements usefull utilities
//
////////////////////////////////////////////////////////////////
#ifndef _UFSD_MISC_H
#define _UFSD_MISC_H

#ifndef IN
# define IN
# define OUT
#endif

///////////////////////////////////////////////////////////
// GetMount
//
// Returns corresponded part of string from file "/proc/mounts"
// Or NULL if device is not mounted
///////////////////////////////////////////////////////////
const char*
GetMount(
    IN const char* szDevice,
    IN bool        bVerbose
    );

///////////////////////////////////////////////////////////
// SplitPath
//
// szPath - absolute path.
//   (1)       Can be represented as <device_name>/<relative_path>
//   (2)       Or                    <mount_name>/<relative_path>
//
// bMounted is set to:
//    0 if name has a form (1)
// or 1 if name has a form (2)
//
//
// If the function successed it returns the pointer into szPath
// which points to relative path and begins with '/'
//
// Returns NULL if not possible to parse
///////////////////////////////////////////////////////////
const char*
SplitPath(
    IN const char*  szPath,
    IN bool         bVerbose,
    OUT bool*       bMounted       // Can be NULL
    );

///////////////////////////////////////////////////////////
// ParsePath
//
//
// szPath - absolute path.
//   (1)       Can be represented as <device_name>/<relative_path>
//   (2)       Or                    <mount_name>/<relative_path>
//
// bMounted is set to:
//    0 if name has a form (1)
// or 1 if name has a form (2)
//
//
// If the function successed it returns the pointer into szPath
// which points to relative path and begins with '/'
//
// Returns NULL if not possible to parse
///////////////////////////////////////////////////////////
const char*
ParsePath(
    IN  const char* szPath,
    IN  bool        bVerbose   = false,
    OUT char**      szDevice   = NULL,
    OUT char**      szMntDir   = NULL,
    OUT char**      szMntOpt   = NULL,
    OUT char**      szMntType  = NULL,
    OUT char**      szRelative = NULL,
    OUT bool*       bMounted   = NULL
    );


///////////////////////////////////////////////////////////
// GetAbsPath
//
// Translates relative path to absolute one
// Returns true if OK
// A-la realpath
///////////////////////////////////////////////////////////
bool
GetAbsPath(
    IN  const char* szRelative,
    IN  bool        bVerbose,
    OUT char*       szAbsPath,
    IN  size_t      nMaxCharsInAbsPath
    );

#endif // _UFSD_MISC_H
