////////////////////////////////////////////////////////////////
// August of 2006. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file contains the code that:
// - Enumerates all directories and files
//
////////////////////////////////////////////////////////////////

#include "ufsd.h"
using namespace UFSD;

#include "uenum.h"

///////////////////////////////////////////////////////////
// CUfsdEnum::EnumDir
//
// This function enumerates directories via UFSD
///////////////////////////////////////////////////////////
unsigned
CUfsdEnum::EnumDir(
    IN CDir*  Dir
    )
{
  unsigned Status       = ERR_NOERROR;
  CEntryNumerator* Enum = NULL;

  if ( NULL == Dir )
  {
    Dir = GetFileSystem()->m_RootDir;
    if ( UFSD_SUCCESS( Dir->GetObjectInfo( m_FileInfo ) ) )
    {
      Status = ProcessObject( Dir, m_FileInfo );
      //
      // break cycle if error occurs
      //
      if ( !UFSD_SUCCESS(Status) && !m_bSkipError )
        return Status;
    }
  }

  //
  // Step 1: Enumerate all files/directories in current directory
  //
  Status = Dir->StartFind( StrASCII, "*", 1, Enum, UFSD_ENUM_SHOW_META );
  if ( ERR_NOERROR == Status )
  {
    while( UFSD_SUCCESS( Status = Dir->FindNext( Enum, m_FileInfo ) ) )
    {
      if ( StrASCII == m_FileInfo.NameType )
      {
        const char* ptr = reinterpret_cast<const char*>(m_FileInfo.Name);
        if ( '\0' == ptr['.' != ptr[0]? 0 : '.' != ptr[1]? 1 : 2] )
          continue;
      }
      else if ( StrUNICODE == m_FileInfo.NameType )
      {
        if ( '\0' == m_FileInfo.Name['.' != m_FileInfo.Name[0]? 0 : '.' != m_FileInfo.Name[1]? 1 : 2] )
          continue;
      }
      else
      {
        continue;
      }

      //
      // Step 2: Open each file/directory
      //
      CFSObject* Fso = NULL;
      Status = Dir->Open( m_FileInfo.NameType, m_FileInfo.Name, m_FileInfo.NameLen, &Fso, &m_FileInfo );
      if ( UFSD_SUCCESS(Status) )
      {
        //
        // Step 3: Call virtual function
        //
        Status = ProcessObject( Fso, m_FileInfo );

        if ( UFSD_SUCCESS( Status )
          && m_bRecursive
          && CFSObject::FSObjDir == Fso->GetObjectType() )
        {
          //
          // Step 4: Call itself recursively
          //
//          assert( FlagOn( m_FileInfo.Attrib, UFSD_SUBDIR ) );
          Status = EnumDir( reinterpret_cast<CDir*>(Fso) );
        }

        //
        // Step 5: Close opened object
        //
        Fso->Destroy();
      }

      //
      // break cycle if error occurs
      //
      if ( !UFSD_SUCCESS(Status) && !m_bSkipError )
        break;

      //
      // Step 6: Continue enumeration
      //
    }
  }

  //
  // Step 7: Free resources
  //
  Enum->Destroy();

  return ERR_NOFILEEXISTS == Status? ERR_NOERROR : Status;
}

