////////////////////////////////////////////////////////////////
// January of 2009. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file includes other files to make one big program
// Used to compile static "all in one program"
////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <stdlib.h> // exit
#include <string.h> // memset
#include <assert.h>
#include <stdarg.h>
#include <memory.h>
#include <ctype.h>

#ifdef _WIN32
//---------------------------------------------------------
// To compile windows utility
//---------------------------------------------------------

# include <windows.h>
#include <io.h>       // _get_osfhandle
# include "./win/misc.h"
# define snprintf   _snprintf
# define strcasecmp stricmp

#undef MessageBox

//
// MSVC has a nice feature that simplifies makefile
//
#pragma comment( lib, "./win/ufsd.lib")

#else
//---------------------------------------------------------
// To compile linux utility
//---------------------------------------------------------
#include <linux/hdreg.h>
#include <unistd.h>
#include <fcntl.h>
#include <locale.h>
#include <sys/ioctl.h>

//
// When this file is compiled as a Linux utility
// then some useful functions (GetAbsPath,ParsePath) that are shared for
// many projects are placed in file misc.cpp
//
# include "misc.h"

//---------------------------------------------------------
// End of Linux specific
//---------------------------------------------------------
#endif

//
// Include main ufsd header
//
#include "ufsd.h"

//
// All UFSD classes and functions are placed in UFSD namespace
//
using namespace UFSD;

//
// This file contains declarations of useful UFSD-client functions
//
#include "funcs.h"
#include "gettext.h"

#ifdef INCLUDE_UFSD_CHKNTFS
namespace CHKNTFS{
  #include "chkntfs.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_INFNTFS
namespace INFNTFS{
  #include "infntfs.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_MKNTFS
namespace MKNTFS{
  #include "mkntfs.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_DFNTFS
namespace DFNTFS{
  #include "dfntfs.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_WIPE
namespace WIPE{
  #include "wipe.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_EXTENTS
namespace EXTENTS{
  #include "extents.cpp"
  #undef NLS_CAT_NAME
  #include "uenum.cpp"
}
#endif

#ifdef INCLUDE_UFSD_FSUTIL
namespace FSUTIL{
  #include "fsutil.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_CHKHFS
namespace CHKHFS{
  #include "chkhfs.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_MKHFS
namespace MKHFS{
  #include "mkhfs.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_CHKEXFAT
namespace CHKEXFAT{
  #include "chkexfat.cpp"
  #undef NLS_CAT_NAME
}
#endif

#ifdef INCLUDE_UFSD_MKEXFAT
namespace MKEXFAT{
  #include "mkexfat.cpp"
  #undef NLS_CAT_NAME
}
#endif


///////////////////////////////////////////////////////////
// main
//
//
///////////////////////////////////////////////////////////
int __cdecl
main(
    IN int         argc,
    IN const char* argv[]
    )
{
  if ( argc < 1 )
    return -1;  // Impossible

  const char* szErr = CheckUFSDEndian();
  if ( NULL != szErr )
  {
    fprintf( stderr, "%s\n", szErr );
    return -1;
  }

  if ( argc >= 2 )
  {
#ifdef INCLUDE_UFSD_CHKNTFS
    if ( 0 == strcasecmp( argv[1], "chkntfs" ) )
      return CHKNTFS::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_INFNTFS
    if ( 0 == strcasecmp( argv[1], "infntfs" ) )
      return INFNTFS::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_MKNTFS
    if ( 0 == strcasecmp( argv[1], "mkntfs" ) )
      return MKNTFS::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_DFNTFS
    if ( 0 == strcasecmp( argv[1], "dfntfs" ) )
      return DFNTFS::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_WIPE
    if ( 0 == strcasecmp( argv[1], "wipe" ) )
      return WIPE::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_EXTENTS
    if ( 0 == strcasecmp( argv[1], "extents" ) )
      return EXTENTS::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_FSUTIL
    if ( 0 == strcasecmp( argv[1], "fsutil" ) )
      return FSUTIL::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_CHKHFS
    if ( 0 == strcasecmp( argv[1], "chkhfs" ) )
      return CHKHFS::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_MKHFS
    if ( 0 == strcasecmp( argv[1], "mkhfs" ) )
      return MKHFS::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_CHKEXFAT
    if ( 0 == strcasecmp( argv[1], "chkexfat" ) )
      return CHKEXFAT::main( argc - 1, argv + 1 );
#endif
#ifdef INCLUDE_UFSD_MKEXFAT
    if ( 0 == strcasecmp( argv[1], "mkexfat" ) )
      return MKEXFAT::main( argc - 1, argv + 1 );
#endif
  }

  fprintf( stdout,
"Usage %s "
#ifdef INCLUDE_UFSD_CHKNTFS
" [chkntfs]"
#endif
#ifdef INCLUDE_UFSD_INFNTFS
" [infntfs]"
#endif
#ifdef INCLUDE_UFSD_MKNTFS
" [mkntfs]"
#endif
#ifdef INCLUDE_UFSD_DFNTFS
" [dfntfs]"
#endif
#ifdef INCLUDE_UFSD_WIPE
" [wipe]"
#endif
#ifdef INCLUDE_UFSD_EXTENTS
" [extents]"
#endif
#ifdef INCLUDE_UFSD_FSUTIL
" [fsutil]"
#endif
#ifdef INCLUDE_UFSD_CHKHFS
" [chkhfs]"
#endif
#ifdef INCLUDE_UFSD_MKHFS
" [mkhfs]"
#endif
#ifdef INCLUDE_UFSD_CHKEXFAT
" [chkexfat]"
#endif
#ifdef INCLUDE_UFSD_MKEXFAT
" [mkexfat]"
#endif
"\n"
           , argv[0] );
  return -2;
}
