////////////////////////////////////////////////////////////////
// October of 2004. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file implements CRWBlock for UFSD library
//
////////////////////////////////////////////////////////////////
#ifdef __QNX__
#define _FILE_OFFSET_BITS 64
#ifdef __QNXNTO__
#include <sys/cam_device.h>
#include <sys/dcmd_cam.h>
#include <sys/dcmd_blk.h>
#endif
#endif

#include <stdio.h>
#if defined __APPLE__ || defined __QNX__
  #include <stdlib.h>
  #include <sys/disk.h>
#else
  #include <malloc.h>
#endif
#include <errno.h>
#include <assert.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifndef __USE_LARGEFILE64
  #define __USE_LARGEFILE64
#endif
#ifndef __USE_FILEOFFSET64
  #define __USE_FILEOFFSET64
#endif
#ifndef __USE_UNIX98
  #define __USE_UNIX98
#endif
//#undef __STDC__
//#define __STDC__ 0

#include <fcntl.h>

#if defined(_WIN32)
//
// Windows specific headers and defines are declared here
//
#include <io.h>

  #define open64 _open
  #define close _close
  #define lseek64 _lseeki64
  #define __off64_t_defined
  static inline unsigned int pread64(int h, void *b, unsigned int c, __int64 off) {
    _lseeki64(h, off, 0);
    return _read(h, b, c);
  }
  static inline unsigned int pwrite64(int h, void *b, unsigned int c, __int64 off) {
    _lseeki64(h, off, 0);
    return _write(h, b, c);
  }
  #define R_OK  4

  #ifndef WIN32_LEAN_AND_MEAN
  #define WIN32_LEAN_AND_MEAN
  #endif
  #include <windows.h>
  #pragma warning( disable: 4201 ) // nonstandard extension used : nameless struct/union
  #include <winioctl.h>

  #ifndef IOCTL_DISK_GET_PARTITION_INFO_EX

    #define IOCTL_DISK_GET_PARTITION_INFO_EX    CTL_CODE(IOCTL_DISK_BASE, 0x0012, METHOD_BUFFERED, FILE_ANY_ACCESS)

    typedef enum _PARTITION_STYLE {
        PARTITION_STYLE_MBR,
        PARTITION_STYLE_GPT,
        PARTITION_STYLE_RAW
    } PARTITION_STYLE;

    typedef struct _PARTITION_INFORMATION_MBR {
        BYTE  PartitionType;
        BOOLEAN BootIndicator;
        BOOLEAN RecognizedPartition;
        DWORD HiddenSectors;
    } PARTITION_INFORMATION_MBR, *PPARTITION_INFORMATION_MBR;


    //
    // The following structure defines information in a GPT partition that is
    // not common to both GPT and MBR partitions.
    //

    typedef struct _PARTITION_INFORMATION_GPT {
        GUID PartitionType;                 // Partition type. See table 16-3.
        GUID PartitionId;                   // Unique GUID for this partition.
        DWORD64 Attributes;                 // See table 16-4.
        WCHAR Name [36];                    // Partition Name in Unicode.
    } PARTITION_INFORMATION_GPT, *PPARTITION_INFORMATION_GPT;


    typedef struct _PARTITION_INFORMATION_EX {
        PARTITION_STYLE PartitionStyle;
        LARGE_INTEGER StartingOffset;
        LARGE_INTEGER PartitionLength;
        DWORD PartitionNumber;
        BOOLEAN RewritePartition;
        union {
            PARTITION_INFORMATION_MBR Mbr;
            PARTITION_INFORMATION_GPT Gpt;
        };
    } PARTITION_INFORMATION_EX, *PPARTITION_INFORMATION_EX;

  #endif

#ifdef _CONSOLE
  #define Trace(a) fprintf a
#endif

#else
  //
  // Linux specific headers and defines are declared here
  //
  #include <string.h> // strerror
  #include <unistd.h>
  #ifdef __APPLE__
    #undef BLKSSZGET
    #define BLKSSZGET DKIOCGETBLOCKSIZE
  #else
    #include <malloc.h>
    #ifndef __QNX__
      #include <linux/hdreg.h>
    #endif
  #endif
  #include <sys/ioctl.h>

  #if defined(_IO) && !defined(BLKPBSZGET)
    #define BLKPBSZGET _IO(0x12,123)
  #endif

  #if defined(_IO) && !defined(BLKSSZGET)
    #define BLKSSZGET  _IO(0x12,104)    // get block device sector size
  #endif

  #if defined(_IO) && !defined(BLKDISCARDZEROES)
    #define BLKDISCARDZEROES _IO(0x12,124)
  #endif

  #if defined(_IO) && !defined(BLKDISCARD)
    #define BLKDISCARD    _IO(0x12,119)
  #endif

  #define Trace(a) fprintf a

#endif

#include "ufsd.h"

#ifndef Trace
  #define Trace(a) do{}while((void)0,0)
#endif

#ifndef IN
# define IN
# define OUT
#endif

// Add missing defines.
#ifndef O_BINARY
  #define O_BINARY 0
#endif
#ifndef O_LARGEFILE
  #define O_LARGEFILE 0
#endif

#ifndef S_IRUSR
  #define S_IRUSR 0
  #define S_IWUSR 0
#endif
#ifndef S_IRGRP
  #define S_IRGRP 0
  #define S_IWGRP 0
#endif

#ifndef __ANDROID__
# ifndef __off64_t_defined
#  define open64 open
#  define lseek64 lseek
#  define pread64 pread
#  define pwrite64 pwrite
# endif
#else
static inline unsigned int pread64(int h, void *b, unsigned int c, long long off) {
  lseek64(h, off, SEEK_SET);
  return read(h, b, c);
}
static inline unsigned int pwrite64(int h, void *b, unsigned int c, long long off) {
  lseek64(h, off, SEEK_SET);
  return write(h, b, c);
}
# define open64 open
#endif

#ifndef max
 #define max(a,b) ((a)>(b)?(a):(b))
#endif

#ifndef min
 #define min(a,b) ((a)<(b)?(a):(b))
#endif

#if defined UFSD_TRACE

  #ifdef __GNUC__
    extern "C" void _UFSDTrace( const char* fmt, ... ) __attribute__ ((format (printf, 1, 2)));
  #else
    // This function must be placed somewhere in the client code
    extern "C" void _UFSDTrace( const char* fmt, ... );
  #endif
  #define UFSDTrace(a) _UFSDTrace a

#else

  #define UFSDTrace(a) do{}while((void)0,0)

#endif


//=============================================================================
//                        CUFSD_RWBlock
//=============================================================================
struct CUFSD_RWBlock : public UFSD::CRWBlock, public ufsd_noncopyable
{
  int               m_hFile;
  bool              m_bReadOnly;
#ifdef WIN32
  bool              m_bLocked;
#else
  bool              m_bNoCache;     // Do not cache data
  bool              m_bZeroes;      // true if volume supports discard+zero request
  bool              m_bWarnPrinted; // true if warning is printed about zeroing
  bool              m_bBlockDevice; // true for block devices
#endif
  bool              m_bVerbose;     // verbose actions

  unsigned int      m_SectorShift;  // BytesPerSector = (1 << m_SectorShift) bytes
  unsigned int      m_BlockShift;   // BlockSize = (1 << m_BlockShift) bytes
  UINT64            m_Size;         // In bytes
  char*             m_szDevice;     // copy of device name

  enum MODE { MODE_READ, MODE_WRITE, MODE_VERIFY };

  // Helper function
  unsigned ReadWriteBlocks(
      IN const UINT64&  BlckNum,
      IN OUT void *     pBuffer,
      IN size_t         BlckCount,
      IN MODE           Mode
      );

  //=============================================
  //    CRWBlock virtual functions
  //=============================================
  virtual unsigned ReadBlocks(
      IN const UINT64& BlckNum,
      IN OUT void *    pBuffer,
      IN size_t        BlckCount
      )
  {
//    Trace(( stderr, "ReadBlocks 0x%" PLL "x : 0x%x\n", BlckNum, BlckCount ));
    // Call helper function
    return ReadWriteBlocks( BlckNum, pBuffer, BlckCount, MODE_READ );
  }

  virtual unsigned WriteBlocks(
      IN const UINT64& BlckNum,
      IN const void *  pBuffer,
      IN size_t        BlckCount
      )
  {
//    Trace(( stderr, "WriteBlocks 0x%" PLL "x : 0x%x\n", BlckNum, BlckCount ));
    if ( m_bReadOnly )
      return ERR_WPROTECT;

    // Call helper function
    return ReadWriteBlocks( BlckNum, const_cast<void*>(pBuffer), BlckCount, MODE_WRITE );
  }

  virtual unsigned int GetBlockSize(){ return 1U << m_BlockShift; }

  virtual unsigned int GetSectorSize() const{ return 1U << m_SectorShift; }

  // Discard range of bytes (known as trim for SSD)
  virtual unsigned DiscardRange(
      IN  const UINT64& Offset,
      IN  const UINT64& Bytes
      )
  {
//    Trace(( stderr, "DiscardBlocks 0x%" PLL "x : 0x%x\n", BlckNum, BlckCount ));
    if ( m_bReadOnly )
      return ERR_WPROTECT;

#ifdef WIN32
    UNREFERENCED_PARAMETER( Offset );
    UNREFERENCED_PARAMETER( Bytes );
#else

#ifdef BLKDISCARD
    UINT64 range[2];
    range[0] = Offset;
    range[1] = Bytes;
    if ( 0 == ioctl( m_hFile, BLKDISCARD, &range ) )
    {
      if ( !m_bZeroes && m_bVerbose && !m_bWarnPrinted )
      {
        Trace(( stderr, "info: device \"%s\" does not zeroes while discarding\n", m_szDevice ));
        m_bWarnPrinted = true;
      }
      return ERR_NOERROR;
    }

    int ret = errno;
    if ( EOPNOTSUPP != ret && ENOTTY != ret ) {
      Trace(( stderr, "ioctl( \"%s\", BLKDISCARD ) failed %d: \"%s\"\n", m_szDevice, ret, strerror( ret ) ));
    }
    return ret;
#else
    UNREFERENCED_PARAMETER( Offset );
    UNREFERENCED_PARAMETER( Bytes );
#endif
#endif
    return ERR_NOTIMPLEMENTED;
  }

  virtual void SetBlockSize( IN unsigned int NewBlockSize ){
    if ( (1U << m_BlockShift) != NewBlockSize )
    {
      m_BlockShift  = UFSD::GetLog2OfBlockSize( NewBlockSize );
      assert( (1u << m_BlockShift) == NewBlockSize );
    }
  }

  virtual UINT64 GetNumberOfBlocks(){
    return m_Size >> m_BlockShift;
  }

  virtual void SetVFSContext(IN void*){};

  // returns non zero for read-only mode
  virtual int IsReadOnly() const { return m_bReadOnly; }

  // Verifies disk. ( A-la surface test )
  // Returns 0 if block is OK
  virtual unsigned Verify(
      IN const UINT64& Blck,
      IN size_t        BlckCount
      )
  {
    // Call helper function
    return ReadWriteBlocks( Blck, NULL, BlckCount, MODE_VERIFY );
  }

  // flush cache buffers (sensible for cache RWBlock only)
  // returns  zero if OK, else UFSD error code.
  virtual unsigned Flush() { return 0; }//fflush( m_hFile );

  //=============================================
  //   End of CRWBlock virtual functions
  //=============================================


  //
  // Constructor is used to initialize
  // class defaults.
  //

  CUFSD_RWBlock( IN bool bReadOnly )
    : m_hFile(-1)
    , m_bReadOnly(bReadOnly)
#ifdef WIN32
    , m_bLocked(false)
#else
    , m_bNoCache(true)
    , m_bZeroes(false)
    , m_bWarnPrinted(false)
    , m_bBlockDevice(true)
    , m_bVerbose(false)
#endif
    , m_SectorShift(9)
    , m_BlockShift(9) // Log2(512)
    , m_Size(0)
    , m_szDevice(NULL)
  {
  }

  virtual ~CUFSD_RWBlock()
  {
    if ( -1 != m_hFile )
    {
#ifdef _WIN32
      if ( m_bLocked )
      {
        DWORD Tmp;
        HANDLE hFile = (HANDLE)::_get_osfhandle(m_hFile);
        // dismount the volume
        ::DeviceIoControl( hFile, FSCTL_DISMOUNT_VOLUME, NULL, 0, NULL, 0, &Tmp, NULL );
        // unlock the volume
        // Really this operation is not required 'cause close( m_hFile )
        ::DeviceIoControl( hFile, FSCTL_UNLOCK_VOLUME, NULL, 0, NULL, 0, &Tmp, NULL );
      }
#endif
      close( m_hFile );
    }
    free( m_szDevice );
  }


  //
  // Init CUFSD_RWBlock
  //
  unsigned Init(
      IN const char* szDevice,
      IN bool        bForceDismount,
      IN bool        bVerbose,
      OUT int*       fd
      )
  {
    m_bVerbose = bVerbose;
#if defined _WIN32 & defined _CONSOLE
    int Attempts = 0;
Again:
#endif
    m_hFile = open64( szDevice,
                      O_LARGEFILE | O_BINARY | (m_bReadOnly? O_RDONLY : O_RDWR),
                      S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP );
    if ( -1 == m_hFile )
    {
      int err = errno;
      Trace(( stderr, "Can't open \"%s\" : %s\n", szDevice, strerror(err) ));
      return err;
    }

#ifndef __APPLE__
    // Get the size of file
    lseek64(m_hFile, 0, SEEK_SET);
    m_Size = lseek64(m_hFile, 0, SEEK_END);
#else
    ioctl(m_hFile, DKIOCGETBLOCKCOUNT, &m_Size);
#endif

#ifdef _WIN32
    //
    // Windows specific part of code
    //
    {
      DWORD Tmp;
      HANDLE hDevice = (HANDLE)::_get_osfhandle(m_hFile);
      unsigned Status = ERR_BADPARAMS;

      DISK_GEOMETRY Geom;
      if ( !::DeviceIoControl( hDevice, IOCTL_DISK_GET_DRIVE_GEOMETRY, NULL, 0, &Geom, sizeof(Geom), &Tmp, NULL ) )
      {
        Trace(( stderr, "Can't get sector size of \"%s\". Use default 512 bytes\n", szDevice ));
        assert( 9 == m_SectorShift );
      }
      else
      {
        if ( bVerbose )
          Trace(( stderr, "DeviceIoControl( \"%s\", IOCTL_DISK_GET_DRIVE_GEOMETRY ) => BytesPerSector = %d\n", szDevice, Geom.BytesPerSector ));

        m_SectorShift = 9 + UFSD::Log2( Geom.BytesPerSector >> 9 );
      }

      //
      // Standard sequence of operation on Raw disks:
      // LOCK + operation + DISMOUNT + UNLOCK
      //
      ::FlushFileBuffers( hDevice );
      if ( ::DeviceIoControl( hDevice, FSCTL_LOCK_VOLUME, NULL, 0, NULL, 0, &Tmp, NULL ) )
      {
        m_bLocked = true;
        if ( bVerbose )
          Trace(( stderr, "\"%s\" is locked\n", szDevice ));
      }
      else if ( !m_bReadOnly )
      {
        Trace(( stderr, "Can't lock volume \"%s\"\n", szDevice ));
#ifdef _CONSOLE
        if ( !bForceDismount )
        {
          fprintf( stderr, "Press 'y' to force dismount \"%s\"",szDevice );
          // Get the first char
          int r = fgetc( stdin );
          // Skip the next until the end
          while( !feof( stdin ) && '\n' != fgetc( stdin ) )
            ;

          if ( 'y' != r )
          {
            Status = ERR_CANCELLED;
            goto Error;
          }
        }

        if ( !::DeviceIoControl( hDevice, FSCTL_DISMOUNT_VOLUME, NULL, 0, NULL, 0, &Tmp, NULL ) )
          Trace(( stderr, "Can't dismount volume \"%s\"\n", szDevice ));
        else if ( ++Attempts < 3 )
        {
          // Reopen volume cause 'hDevice' becomes invalid
          if ( bVerbose )
            Trace(( stderr, "force dismount: reopen \"%s\"\n", szDevice ));
          close( m_hFile );
          Sleep( 2000 );
          goto Again;
        }
#else
        UNREFERENCED_PARAMETER( bForceDismount );
#endif

Error:
        close( m_hFile );
        m_hFile = -1;
        return Status;
      }

      if ( (UINT64)-1 == m_Size )
      {
        union {
          PARTITION_INFORMATION_EX ex;
          PARTITION_INFORMATION mbr;
        } Info;
        const char *hint = NULL;
        // Get the size of volume
        if ( ::DeviceIoControl( hDevice, IOCTL_DISK_GET_PARTITION_INFO_EX,
                                NULL, 0,
                                &Info.ex, sizeof(Info.ex),
                                &Tmp, NULL ) )
        {
          m_Size = Info.ex.PartitionLength.QuadPart;
          hint = "IOCTL_DISK_GET_PARTITION_INFO_EX";
        }
        else if ( ::DeviceIoControl( hDevice, IOCTL_DISK_GET_PARTITION_INFO,
                                     NULL, 0,
                                    &Info.mbr, sizeof(Info.mbr),
                                    &Tmp, NULL ) )
        {
          m_Size = Info.mbr.PartitionLength.QuadPart;
          hint = "IOCTL_DISK_GET_PARTITION_INFO";
        }
        else
        {
          Trace(( stderr, "Can't get the size of volume \"%s\"\n", szDevice ));
          goto Error;
        }

        if ( bVerbose )
          Trace(( stderr, "DeviceIoControl( \"%s\", %s ) => 0x%" PLL "x\n", szDevice, hint, m_Size ));
      }

      // Allow to write to the end of volume
#ifndef FSCTL_ALLOW_EXTENDED_DASD_IO
  #define FSCTL_ALLOW_EXTENDED_DASD_IO  0x00090083
#endif
      if ( !::DeviceIoControl( hDevice, FSCTL_ALLOW_EXTENDED_DASD_IO,
                               NULL, 0, NULL, 0, &Tmp, NULL )
        && ERROR_INVALID_PARAMETER != GetLastError() )
      {
        Trace(( stderr, "FSCTL_ALLOW_EXTENDED_DASD_IO failed \"%s\"\n", szDevice ));
      }
      else
      {
        if ( bVerbose )
          Trace(( stderr, "\"%s\": allowed EXTENDED_DASD_IO\n", szDevice ));
      }
    }
#else
    //
    // Linux specific part of code
    //
    struct stat st;
    if ( 0 == fstat( m_hFile, &st ) )
    {
      if ( bVerbose )
        Trace(( stderr, "fstat(\"%s\") returns mode %o\n", szDevice, (unsigned)st.st_mode ));

      if ( !S_ISBLK( st.st_mode ) )
      {
        m_bBlockDevice  = false;
      }
    }
    else
    {
      Trace(( stderr, "fstat(\"%s\") failed \"%s\"\n", szDevice, strerror(errno) ));
    }

    if ( m_bBlockDevice )
    {
      // Get the size of sector in bytes
      int BytesPerSector = 0;

#ifdef __QNX__
    // Qnx specific
    #ifdef __QNXNTO__
      cam_devinfo_t dbuf = { 0 };
      int r = devctl( m_hFile, DCMD_CAM_DEVINFO, &dbuf, sizeof( dbuf ), NULL );
      BytesPerSector = dbuf.sctr_size;
      if ( r != EOK )
      {
        Trace(( stderr, "devctl(DCMD_CAM_DEVINFO) failed (%s). Use default sector size(512)\n", strerror(errno) ));
        BytesPerSector = 0;
      }
    #else
      BytesPerSector = QNX4FS_BLOCK_SIZE;
      if ( 0 )
      {
          ;
      }
    #endif
#else
      // BLKSSZGET: get block device logical block size
      // or
      // BLKSSZGET: get block device hardware sector size (if BLKPBSZGET is not defined)
      if ( 0 == ioctl( m_hFile, BLKPBSZGET, &BytesPerSector ) )
      {
        if ( bVerbose )
          Trace(( stderr, "ioctl(\"%s\",BLKPBSZGET) returns %d\n", szDevice, BytesPerSector ));
      }
      else
      {
        int err = errno;
        if ( 0 == ioctl( m_hFile, BLKSSZGET, &BytesPerSector ) )
        {
          Trace(( stderr, "ioctl(BLKPBSZGET) failed (%s). ioctl(BLKSSZGET) returns sector size %d\n", strerror(err), BytesPerSector ));
        }
        else
        {
          Trace(( stderr, "ioctl(BLKPBSZGET/BLKSSZGET) failed (%s). Use default sector size(512)\n", strerror(errno) ));
          BytesPerSector = 512;
        }
      }
#endif
      m_SectorShift = 9 + UFSD::Log2( (unsigned)BytesPerSector >> 9 );
    }

#ifdef __APPLE__
    m_Size <<= m_SectorShift;
#endif

#if defined BLKDISCARDZEROES && !defined __QNX__
    //
    // Check for discard + zeroes (Really we need this code only if Verbose is true.
    //
    {
      int zeroes = 0;
      if ( 0 != ioctl( m_hFile, BLKDISCARDZEROES, &zeroes ) )
      {
        m_bZeroes = false;
        if ( bVerbose )
          Trace(( stderr, "ioctl(\"%s\", BLKDISCARDZEROES) failed (%s)\n", szDevice, strerror(errno) ));
      }
      else if ( 0 == zeroes )
      {
        m_bZeroes = false;
        if ( bVerbose )
          Trace(( stderr, "\"%s\": discardzeroes is not supported by this device\n", szDevice ));
      }
      else
      {
        if ( bVerbose )
          Trace(( stderr, "\"%s\": discardzeroes is supported\n", szDevice ));
        m_bZeroes = true;
      }
    }
#endif
#endif // #ifdef _WIN32

    // Rewind back
    //lseek64(m_hFile, 0ULL, SEEK_SET);

    if ( NULL != fd )
      *fd = m_hFile;

    switch( m_SectorShift )
    {
    case 9:   break;
    case 12:
      Trace(( stderr, "%s: 4K sector size\n", szDevice ));
      break;
    default:
      Trace(( stderr, "%s: non standard sector size 0x%x bytes\n", szDevice,  1u << m_SectorShift ));
    }

    if ( bVerbose )
      Trace(( stderr, "\"%s\": disk size 0x%" PLL "x bytes. ~%" PLL "uGB\n", szDevice, m_Size, (m_Size + 0x20000000) >> 30 ));

    // By default block equals sector
    m_BlockShift = m_SectorShift;

    // Make a copy of device name to check for enjected media
    m_szDevice = strdup( szDevice );

    return ERR_NOERROR;
  }
};



///////////////////////////////////////////////////////////
// CUFSD_RWBlock::ReadWriteBlocks
//
// This helper function is called from Read/Write/Verify functions
///////////////////////////////////////////////////////////
unsigned
CUFSD_RWBlock::ReadWriteBlocks(
    IN const UINT64& BlckNum,
    IN OUT void *    pBuffer,
    IN size_t        BlckCount,
    IN MODE          Mode
    )
{
  if ( -1 == m_hFile )
    return ERR_BADPARAMS;

  if ( 0 == BlckCount )
    return ERR_NOERROR;

  // Translate blocks to bytes
  UINT64 Offset = BlckNum << m_BlockShift;

  // Check boundary
  if( Offset + (BlckCount<<m_BlockShift) > m_Size )
    return ERR_BADPARAMS;

  size_t ToProcess = BlckCount << m_BlockShift;
  unsigned Processed;

  switch( Mode )
  {
  case MODE_READ:
    Processed = pread64( m_hFile, pBuffer, ToProcess, Offset );
    if ( (unsigned)-1 == Processed
       || Processed   != ToProcess )
    {
      Trace(( stderr, "Error reading sector 0x%" PLL "x, count 0x%x, retval=%d, error=%s\n",
              Offset>>m_SectorShift, (unsigned)(ToProcess>>m_SectorShift), (int)Processed, strerror(errno) ));
      if ( NULL != m_szDevice && access( m_szDevice, R_OK ) )
        return ERR_NOMEDIA;

      return ERR_READFILE;
    }
    break;

  case MODE_WRITE:
    Processed = pwrite64( m_hFile, pBuffer, ToProcess, Offset );
    if ( (unsigned)-1 == Processed
       || Processed   != ToProcess )
    {
      Trace(( stderr, "Error writing sector 0x%" PLL "x, count 0x%x, retval=%d, error=%s\n",
              Offset>>m_SectorShift, (unsigned)(ToProcess>>m_SectorShift), (int)Processed, strerror(errno) ));
      if ( NULL != m_szDevice && access( m_szDevice, R_OK ) )
        return ERR_NOMEDIA;
      return ERR_WRITEFILE;
    }
    break;

  case MODE_VERIFY:
#ifdef _WIN32
    {
      DWORD Tmp;
      VERIFY_INFORMATION Info;
      Info.StartingOffset.QuadPart = Offset;
      Info.Length         = (DWORD)ToProcess;

      if ( !::DeviceIoControl( (HANDLE)::_get_osfhandle(m_hFile), IOCTL_DISK_VERIFY,
                               &Info, sizeof(Info), NULL, 0, &Tmp, NULL ) )
      {
        return ERR_BADPARAMS;
      }
    }
#else
    {
      // I don't know how to verify disks in Linux
      // Just try to read
      // Use 128Kb blocks
      size_t BlockSize = min( 128*1024, ToProcess );
      unsigned char* Tmp = (unsigned char*)malloc( BlockSize );
      if ( NULL != Tmp )
      {
        while( 0 != ToProcess )
        {
          size_t part = min( ToProcess, BlockSize );
          Processed = pread64( m_hFile, Tmp, part, Offset );
          if ( (unsigned)-1 == Processed
            || Processed   != part )
          {
            free( Tmp );
            return ERR_BADPARAMS;
          }

          ToProcess -= part;
          Offset    += part;
        }
        free( Tmp );
      }
    }
#endif
    break;

  default:
    return ERR_BADPARAMS;
  }

#if 0 //ndef WIN32
  if ( m_bNoCache && posix_fadvise64( m_hFile, 0, 0, POSIX_FADV_DONTNEED ) < 0 )
  {
    static int printed;
    if ( !printed )
    {
      printed = 1;
      Trace(( stderr, "Advise error :%s\n", strerror(errno) ));
    }
  }
#endif

  return ERR_NOERROR;
}


///////////////////////////////////////////////////////////
// UFSD_IOHandlerCreate
//
// The API exported from this module.
// It returns the pointer to new CUFSD_RWBlock
// NOTE: the caller should call UFSD_IOHandlerFree
// to free resources
///////////////////////////////////////////////////////////
unsigned
UFSD_IOHandlerCreate(
    IN const char*       szDevice,
    IN bool              bReadOnly,
    OUT UFSD::CRWBlock** RwBlock,
    OUT int*             fd,
    IN bool              bForceDismount,
    IN bool              bVerbose
   )
{
  // Set the default return value
  *RwBlock = NULL;

  // Try to allocate
  CUFSD_RWBlock* rw = new CUFSD_RWBlock( bReadOnly );
  if ( NULL == rw )
    return ERR_NOMEMORY;

  // Try to init
  unsigned err = rw->Init( szDevice, bForceDismount, bVerbose, fd );

  // Check error
  if ( ERR_NOERROR == err )
    *RwBlock = rw;
  else
    delete rw;

  return err;
}


///////////////////////////////////////////////////////////
// UFSD_IOHandlerFree
//
// The API exported from this module.
///////////////////////////////////////////////////////////
void
UFSD_IOHandlerFree(
    IN UFSD::CRWBlock* p
   )
{
  delete static_cast<CUFSD_RWBlock*>(p);
}
