////////////////////////////////////////////////////////////////
// October of 2004. (C) Paragon Software
// If this code works, it was written by Alexander Mamaev
// If not, I don't know who wrote it.
//
// This file implements CTime for UFSD library
//
////////////////////////////////////////////////////////////////

#include <time.h>
#if defined(_WIN32) && defined(_MSC_VER)
#define timezone _timezone
#define tzset _tzset
#endif

#ifdef __APPLE__
#include <sys/time.h>
#endif

#ifdef __ANDROID__
#include <linux/time.h>
#endif

// Include UFSD specific files
#include "ufsd.h"

#define _100ns2seconds  (10*1000*1000)
#define SecondsToStartOf1970 PU64(0x00000002B6109100)


class UFSD_Time : public UFSD::CTime
{
public:
  // This function returns UTC
  virtual UINT64 Time()
  {
    return (SecondsToStartOf1970 + time(0)) * _100ns2seconds;
  }

  // The bias is the difference, in minutes, between UTC time and local time.
  // F.e.
  //  Eastern time zone: +300
  //  Paris,Berlin:      -60
  //  Moscow:            -180
  virtual long GetBias()
  {
#if !defined __APPLE__ && !defined __ANDROID__
    return timezone/60; // timezone stores it in seconds.
#else
    struct timezone tz = {0};
    if ( 0 == gettimeofday(NULL, &tz) )
      return tz.tz_minuteswest;
    return 0;
#endif
  }
};

// The only Time Manager
static UFSD_Time s_TimeMngr;


///////////////////////////////////////////////////////////
// UFSD_GetTimeService
//
// The API exported from this module.
// It returns the pointer to the static class.
///////////////////////////////////////////////////////////
UFSD::CTime*
UFSD_GetTimeService(void)
{
  tzset();
  return &s_TimeMngr;
}

